import pyvisa as pv
from pylogfile import *
import numpy as np
import time
import inspect

class Identifier:
	
	def __init__(self):
		self.idn_model = "" # Identifier provided by instrument itself (*IDN?)
		self.ctg = "" # Category class of driver
		self.dvr = "" # Driver class
		self.name = "" # Rich name provided by user (optional)

class RemoteInstrument:
	''' Class to represent an instrument driven by another host on this network. This
	class allows remote clients to control the instrument, despite not having a 
	connection or driver locally.
	'''
	
	def __init__(self):
		self.net_id = "RohdeSchwarz-FSQ"
		self.client_agent = None
	
	def remote_call(self, func_name:str, *args, **kwargs):
		''' Calls the function 'func_name' of a remote instrument '''
		
		arg_str = ""
		for a in args:
			arg_str = arg_str + f"{a} "
		for key, value in kwargs.items():
			arg_str = arg_str + f"{key}:{value} "
		
		print(f"Initializing remote call: function = {func_name}, arguments = {arg_str} ")

def RemoteFunction(func):
	'''Decorator to allow empty functions to call
	their remote counterparts'''
	
	def wrapper(self, *args, **kwargs):
		self.remote_call(func.__name__, *args, **kwargs)
		func(self, *args, **kwargs)
	return wrapper

class SpectrumAnalyzerRemote(RemoteInstrument):
	
	def __init__(self):
		super().__init__()
	
	# Without the decorator, it looks like this
	def set_freq_start(self, f_Hz:float, channel:int=1):
		self.remote_call('set_freq_start', f_Hz, channel)
	
	# With the decorator, it looks like this
	@RemoteFunction
	def set_freq_end(self, f_Hz:float, channel:int=1):
		pass

class Driver:
	
	def __init__(self, address:str, log:LogPile, expected_idn:str=""):
		
		self.address = address
		self.log = log
		
		self.id = Identifier()
		self.expected_idn = expected_idn 
		self.verified_hardware = False
		
		self.online = False
		self.rm = pv.ResourceManager()
		self.isnt = None
		self.connect()
	
	def connect(self, check_id:bool=True):
		
		try:
			self.inst = self.rm.open_resource(self.address)
			self.online = True
			
			if check_id:
				self.query_id()
			
		except Exception as e:
			self.log.error(f"Failed to connect to address: {self.address}. ({e})")
			self.online = False
	
	def preset(self):
		self.write("*RST")
	
	def query_id(self):
		''' Checks the IDN of the instrument, and makes sure it matches up.'''
		
		# Query IDN model
		self.id.idn_model = self.query("*IDN?")
		
		if self.id.idn_model is not None:
			self.online = True
			self.log.debug(f"Instrument connection state: >ONLINE<")
			
			if self.expected_idn is None or self.expected_idn == "":
				self.log.debug("Cannot verify hardware. No verification string provided.")
				return
			
			# Check if model is right
			if self.expected_idn.upper() in self.id.idn_model.upper():
				self.verified_hardware = True
				self.log.debug(f"Hardware verification >PASSED<", detail=f"Received string: {self.id.idn_model}")
			else:
				self.verified_hardware = False
				self.log.debug(f"Hardware verification >FAILED<", detail=f"Received string: {self.id.idn_model}")
		else:
			self.log.debug(f"Instrument connection state: >OFFLINE<")
			self.online = False
		
	def close(self):
		
		self.inst.close()
	
	def wait_ready(self, check_period:float=0.1, timeout_s:float=None):
		''' Waits until all previous SCPI commands have completed. *CLS 
		must have been sent prior to the commands in question.
		
		Set timeout to None for no timeout.
		
		Returns true if operation completed, returns False if timeout occured.'''
		
		self.write(f"*OPC")
		
		# Check ESR
		esr_buffer = int(self.query(f"*ESR?"))
		
		t0 = time.time()
		
		# Loop while ESR bit one is not set
		while esr_buffer == 0:
			
			# Check register state
			esr_buffer = int(self.query(f"*ESR?"))
			
			# Wait prescribed time
			time.sleep(check_period)
			
			# Timeout handling
			if (timeout_s is not None) and (time.time() - t0 >= timeout_s):
				break
		
		# Return
		if esr_buffer > 0:
			return True
		else:
			return False
		
	def write(self, cmd:str):
		''' Sends a SCPI command via PyVISA'''
		
		if not self.online:
			self.log.warning(f"Cannot write when offline. ()")
		
		try:
			self.inst.write(cmd)
		except Exception as e:
			self.log.error(f"Failed to write to instrument {self.address}. ({e})")
			self.online = False
		
	def id_str(self):
		pass
	
	def read(self):
		''' Reads via PyVISA'''
		
		if not self.online:
			self.log.warning(f"Cannot write when offline. ()")
		
		try:
			return self.inst.write()
		except Exception as e:
			self.log.error(f"Failed to read from instrument {self.address}. ({e})")
			self.online = False
			return None
	
	def query(self, cmd:str):
		''' Querys a command via PyVISA'''
		
		if not self.online:
			self.log.warning(f"Cannot write when offline. ()")
		
		try:
			return self.inst.query(cmd)
		except Exception as e:
			self.log.error(f"Failed to query instrument {self.address}. ({e})")
			self.online = False
			return None

def bool_to_str01(val:bool):
	''' Converts a boolean value to 0/1 as a string '''
	
	if val:
		return "1"
	else:
		return "0"

def str01_to_bool(val:str):
	''' Converts the string 0/1 to a boolean '''
	
	if '1' in val:
		return True
	else:
		return False

def bool_to_ONFOFF(val:bool):
	''' Converts a boolean value to 0/1 as a string '''
	
	if val:
		return "ON"
	else:
		return "OFF"

def str_to_bool(val:str):
	''' Converts the string 0/1 or ON/OFF or TRUE/FALSE to a boolean '''
	
	if ('1' in val) or ('ON' in val.upper()) or ('TRUE' in val.upper()):
		return True
	else:
		return False

def s2hms(seconds):
	''' Converts a value in seconds to a tuple of hours, minutes, seconds.'''
	
	# Convert seconds to minutes
	min = np.floor(seconds/60)
	seconds -= min*60
	
	# Convert minutes to hours
	hours = np.floor(min/60)
	min -= hours*60
	
	return (hours, min, seconds)


def interpret_range(rd:dict, print_err=False):
	''' Accepts a dictionary defining a sweep list/range, and returns a list of the values. Returns none
	if the format is invalid.
	
	* Dictionary must contain key 'type' specifying the string 'list' or 'range'.
	* Dictionary must contain a key 'unit' specifying a string with the unit.
	* If type=list, dictionary must contain key 'values' with a list of each value to include.
	* If type=range, dictionary must contain keys start, end, and step each with a float value
	  specifying the iteration conditions for the list.
	
	Example list dict (in JSON format):
		 {
			"type": "list",
			"unit": "dBm",
			"values": [0]
		}
		
	Example range dict (in JSON format):
		{
			"type": "range",
			"unit": "Hz",
			"start": 9.8e9,
			"step": 1e6,
			"end": 10.2e9
		}
	
	'''
	K = rd.keys()
	
	# Verify type parameter
	if "type" not in K:
		if print_err:
			print(f"    {Fore.RED}Key 'type' not present.{Style.RESET_ALL}")
		return None
	elif type(rd['type']) != str:
			if print_err:
				print(f"    {Fore.RED}Key 'type' wrong type.{Style.RESET_ALL}")
			return None
	elif rd['type'] not in ("list", "range"):
		if print_err:
			print(f"    {Fore.RED}Key 'type' corrupt.{Style.RESET_ALL}")
		return None
	
	# Verify unit parameter
	if "unit" not in K:
		if print_err:
			print(f"    {Fore.RED}Key 'unit' not present.{Style.RESET_ALL}")
		return None
	elif type(rd['unit']) != str:
		if print_err:
			print(f"    {Fore.RED}Key 'unit' wrong type.{Style.RESET_ALL}")
		return None
	elif rd['unit'] not in ("dBm", "V", "Hz"):
		if print_err:
			print(f"    {Fore.RED}Key 'unit' corrupt.{Style.RESET_ALL}")
		return None
	
	# Read list type
	if rd['type'] == 'list':
		try:
			vals = rd['values']
		except:
			if print_err:
				print(f"    {Fore.RED}Failed to read value list.{Style.RESET_ALL}")
			return None
	elif rd['type'] == 'range':
		try:
			
			start = int(rd['start']*1e6)
			end = int(rd['end']*1e6)+1
			step = int(rd['step']*1e6)
			
			vals = np.array(range(start, end, step))/1e6
			
			vals = list(vals)
			
		except Exception as e:
			if print_err:
				print(f"    {Fore.RED}Failed to process sweep values. ({e}){Style.RESET_ALL}")
			return None
	
	return vals


from heimdallr.base import *

class SpectrumAnalyzerCtg(Driver):
	
	SWEEP_CONTINUOUS = "sweep-continuous"
	SWEEP_SINGLE = "sweep-single"
	SWEEP_OFF = "sweep-off"
	
	def __init__(self, address:str, log:LogPile, expected_idn:str=""):
		super().__init__(address, log, expected_idn=expected_idn)
	
	def set_freq_start(self, f_Hz:float, channel:int=1):
		pass
	
	def set_freq_end(self, f_Hz:float, channel:int=1):
		pass
	
	def set_num_points(self, points:int, channel:int=1):
		pass
	
	def set_res_bandwidth(self, rbw_Hz:float, channel:int=1):
		pass
	
	def clear_traces(self):
		pass
	
	def add_trace(self, channel:int, measurement:str):
		''' Returns trace number '''
		pass
	
	def get_waveform(self, channel:int):
		pass


from heimdallr.base import *

class OscilloscopeCtg(Driver):
	
	def __init__(self, address:str, log:LogPile, expected_idn=""):
		super().__init__(address, log, expected_idn=expected_idn)
	
	def set_div_time(self, time_s:float):
		pass
	
	def get_div_time(self, channel:int):
		pass
	
	def set_offset_time(self, channel:int, time_s:float):
		pass
	
	def set_div_volt(self, channel:int, volt_V:float):
		pass
	
	def set_offset_volt(self, channel:int, volt_V:float):
		pass
	
	def set_chan_enable(self, channel:int, enable:bool):
		pass
	
	def get_waveform(self, channel:int):
		pass

class Oscilloscope2Ctg(OscilloscopeCtg):
	
	# Measurement options
	MEAS_VMAX = 0
	MEAS_VMIN = 1
	MEAS_VAVG = 2
	MEAS_VPP  = 3
	MEAS_FREQ = 4
	
	# Statistics options for measurement options
	STAT_NONE = 0
	STAT_AVG = 1
	STAT_MAX = 2
	STAT_MIN = 3
	STAT_CURR = 4
	STAT_STD = 5
	
	def __init__(self, address:str, log:LogPile, expected_idn=""):
		super().__init__(address, log, expected_idn=expected_idn)
	
	def add_measurement(self):
		pass
	
	def get_measurement(self):
		pass


import socket
import rsa
import tabulate
import threading
import hashlib
import sqlite3
from Crypto.Random import get_random_bytes
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad
from pylogfile import *

ACCOUNT_ADMIN = 30
ACCOUNT_STANDARD = 20
ACCOUNT_LOW = 10

# Initialize database access
db_mutex = threading.Lock() # Create a mutex for the database

class UserDatabase:
	""" Handles interactions with, and manipulations of the user data database.
	
	It is supposed to abstract two things:
		1.) Hide database access from the coder using the 'db_mutex' variable
		2.) Hide SQL queries by burying them in functions
		3.) Hide database structure (in SQL queries) by burying it in functions.
		
	This way it's easy to restrucutre the database in the future if needed, and 
	there's no need to worry about database resource locks/races.
	"""
	
	def __init__(self, filename:str):
		
		# Name of database file
		self.filename = filename
		
	def remove_user(self, username:str):
		""" Deletes a user from the database """
		
		# Aquire mutex to protect database
		with db_mutex:
			
			conn = sqlite3.connect(self.filename)
			cur = conn.cursor()
			
			cur.execute("DELETE FROM userdata WHERE username = ?", (username,))
			
			conn.commit()
		
	def add_user(self, username:str, password:str, email:str, usr_type:str):
		""" Adds a user to the database """

		# Verify that user type is valid
		if usr_type not in [ACCOUNT_LOW, ACCOUNT_STANDARD, ACCOUNT_ADMIN]:
			usr_type = ACCOUNT_LOW
		
		# Get hash of password
		password_hash = hashlib.sha256(password.encode()).hexdigest()
		
		# Aquire mutex to protect database
		with db_mutex:
		
			conn = sqlite3.connect(self.filename)
			cur = conn.cursor()
						
			# Lookup highest account ID
			cur.execute("SELECT MAX(acct_id) FROM userdata")
			fd = cur.fetchall()
			try:
				next_ID = int(fd[0][0])+1
			except:
				self.log.critical(f"{self.id_str}Failed to access account ID from database.")
				return False
		
			cur.execute("INSERT INTO userdata (username, password, acct_id, email_addr, verified, acct_type) VALUES (?, ?, ?, ?, ?)", (username, password_hash, next_ID , email, "No", usr_type))
			conn.commit()
	
	def get_user_id(self, username:str):
		""" Accepts a username and looks up the ID of that user. Returns None if 
		user is not found in the database."""
		
		# Aquire mutex to protect database
		with db_mutex:
			
			conn = sqlite3.connect(self.filename)
			cur = conn.cursor()
			
			# Check for user
			cur.execute("SELECT id FROM userdata WHERE username = ?", (username,))
			qd = cur.fetchall()
			if qd:
				return qd[0][0]
			else:
				return None
	
	def get_user_type(self, username:str):
		""" Accepts a username and looks up the account type of that user. Returns None if
		user is not found in the database."""
		
		# Aquire mutex to protect database
		with db_mutex:
			
			conn = sqlite3.connect(self.filename)
			cur = conn.cursor()
			
			# Check for user
			cur.execute("SELECT acct_type FROM userdata WHERE username = ?", (username,))
			qd = cur.fetchall()
			if qd:
				return qd[0][0]
			else:
				return None

	def view_database(self):
		""" Access the entire database contents and return a table string"""
		
		# Hardcode - do not show password
		show_password = False
		
		# Query data from database
		with db_mutex:
			conn = sqlite3.connect(self.filename)
			cur = conn.cursor()
			
			cur.execute("SELECT * FROM userdata")
			
			# Get names
			names = list(map(lambda x: x[0], cur.description))
			
			# If told to hide password, deletes entry from names
			del_idx = None
			if not show_password:
				
				# Find password entry
				for idx, n in enumerate(names):
					if n == "password":
						del_idx = idx
				
				# Delete password item
				if del_idx is not None:
					del names[del_idx]
			
			# Get data
			cd = cur.fetchall()
		
		# Initialize master list
		table_data = []
		table_data.append(names)
		
		# Add user entries, one by one
		for entry in cd:
			
			# Create list
			entry_list = []
					
			# Scan over items in user entry
			for idx, e_item in enumerate(entry):
				
				# Skip passwords
				if idx == del_idx:
					continue
				
				entry_list.append(e_item)
		
			# Add to master lis
			table_data.append(entry_list)

		# Create table	
		T = tabulate.tabulate(table_data, headers='firstrow', tablefmt='fancy_grid')
		
		return str(T)

ENC_FALSE = 100
ENC_AUTO = 101
ENC_TRUE = 102

class ClientAgent:
	""" This class handles networking for the client. It communicates with the
	server and handles encryption and login.
	"""
	
	def __init__(self, conf:dict, log:LogPile):
		
		self.log = log
		
		self.sock = None
		self.ipaddr = conf.core_data['server_addr']
		self.port = conf.core_data['server_port']
		self.addr = (self.ipaddr, self.port)

		self.user = None # If logged in, this is the connected user

		# Generate keys
		self.public_key, self.private_key = rsa.newkeys(1024)

		# Server's public key
		self.server_key = None
		
		# AES variables
		self.aes_key = None
		self.aes_iv = None
		
		self.error_code = None # Code from last error. This will also be replaced by reply() replies that begin with 'ERROR:'
		self.reply = "" # string from last query
		
		#------------------------------------------------------------------#
		# These are parameters synced from the server
	
	def connect_socket(self):
		""" Creates a socket and tries to join the server. Must be called
		after the address has already been set with set_addr(). Can be used
		for both initial connections, and reconnecting. """
		
		# Create new socket
		self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		
		# Connect to socket
		try:
			self.sock.connect(self.addr)
		except:
			self.log.warning("Failed to connect to server")
			return

		
		# Perform handshake
		self.handshake()
	
	def err(self):
		"""Returns the last error code. Returns None if last code already read."""
		
		# Get last code, then reset to None
		ec = self.error_code
		self.error_code = None
		
		if ec == None:
			ec = "--"
		
		return ec
	
	def send(self, x, encode_rule=ENC_AUTO):
		""" Encrypts and sends binary data to the server. If a string is provided,
		it will automatically encode the string unless 'no_encode' is set to true.
		Returns pass/fail status.
		
		Uses AES encryption standard.
		"""
		
		cipher = AES.new(self.aes_key, AES.MODE_CBC, iv=self.aes_iv)
		
		# Automatically encode strings
		if encode_rule == ENC_AUTO:
			if isinstance(x, str):
				x = x.encode()
		elif encode_rule == ENC_TRUE:
			x = x.encode()

		# Try to send encrypted message
		try:
			self.sock.send(cipher.encrypt(pad(x, AES.block_size)))
		except socket.error as e:
			self.log.error(f"Failed to send data to server. Closing Connection. ({str(e)})")
			
			# Tell state to reconnect
			self.state = ClientAgent.CS_HAND
			self.user = None
			
			return False

		return True
	
	def send_numlist(self, L:list):
		""" Encodes a numeric list as a string, and sends it to the server. """
		
		s = ""
		for el in L:
			s += f"{el}:"
		
		# Remove last colon
		s = s[:-1]
		
		return self.send(s)
	
	def recv(self):
		"""	Receives and decrypts binary data from the server.
		
		Uses AES encryption standard."""
		
		cipher = AES.new(self.aes_key, AES.MODE_CBC, iv=self.aes_iv)
		
		# Try to receive encrypted message
		try:
			rv = unpad(cipher.decrypt(self.sock.recv(PACKET_SIZE)), AES.block_size)
		except socket.error as e:
			self.log.error(f"Failed to receive data from server. Closing connection.({str(e)})")
			
			# Tell state to reconnect
			self.state = ClientAgent.CS_HAND
			self.user = None
			
			return None
		except Exception as e:
			self.log.error(f"Failed to receive or decrypt message from client. ({str(e)})")
			return None

		return rv
	
	def recv_str(self):
		""" Receives and decrypts a string from the server."""

		data = self.recv()

		if data is None:
			return None

		try:
			return data.decode()
		except Exception as e:
			self.log.error(f"Encountered an error during decoding in recv_str(). Message: {e}")
			return None
	
	def rsa_send(self, x, encode_rule=ENC_AUTO):
		""" NOTE: This uses RSA encryption and should ONLY be used to exchange AES keys
		as it imposes a length limit on the data.
		
		Encrypts and sends binary data to the server. If a string is provided,
		it will automatically encode the string unless 'no_encode' is set to true.
		Returns pass/fail status.
		"""

		# if not self.online:
		# 	self.log.warning("Cannot send while offline.")
		# 	return False

		# Automatically encode strings
		if encode_rule == ENC_AUTO:
			if isinstance(x, str):
				x = x.encode()
		elif encode_rule == ENC_TRUE:
			x = x.encode()

		# Try to send encrypted message
		try:
			self.sock.send(rsa.encrypt(x, self.server_key))
			# self.sock.send(x)
		except socket.error as e:
			self.log.error(f"Failed to send data to server. Closing Connection. ({str(e)})")
			
			# Tell state to reconnect
			self.state = ClientAgent.CS_HAND
			self.user = None
			
			return False

		return True

	def rsa_recv(self):
		""" NOTE: This should only be used to exchange AES keys.
		
		Receives and decrypts binary data from the server."""

		# if not self.online:
		# 	self.log.warning("Cannot receive while offline.")
		# 	return None

		# Try to receive encrypted message
		try:
			rv = rsa.decrypt(self.sock.recv(PACKET_SIZE), self.private_key)
			# rv = self.sock.recv(PACKET_SIZE)
		except socket.error as e:
			self.log.error(f"Failed to receive data from server. Closing connection.({str(e)})")
			
			# Tell state to reconnect
			self.state = ClientAgent.CS_HAND
			self.user = None
			
			return None
		except Exception as e:
			self.log.error(f"Failed to receive or decrypt message from client. ({str(e)})")
			return None

		return rv

	def set_addr(self, ipaddr, port):
		""" Sets the address to which the class will connect when 'login()' is
		called.
		"""

		self.ipaddr = ipaddr
		try:
			self.port = int(port)
			self.addr = (self.ipaddr, self.port)
		except:
			return False

		return True

	def query(self, x):
		"""Sends a message to the server agent and saves the reply string to 
		self.reply. If the reply string is an error (begins with ERRROR:), the 
		error code is saved as an int in self.error_code. """
		
		self.log.debug(f"QUERY(): Sending '{x}'")
		
		# Send encrypted message
		if not self.send(x):
			self.log.warning("Query aborted because send failed.")
			return False
		
		# Receive encrypted message and save to 'reply'
		self.reply = self.recv_str()
		if self.reply is None:
			self.log.error(f"query returned None. (msg: {x})")
			return False
		
		# self.log.debug(f"QUERY(): Received '{self.reply}'")
		
		# Populate reply_ec if error detected
		if self.reply[0:6] == "ERROR:":
			self.error_code = self.reply[6:]

		return True
	
	def query_liststr(self, L:list):
		""" Query, but sends list data. Returns data still as a string. """
		
		self.log.debug(f"QUERY(): Sending '{L}'")
		
		# Send encrypted message
		if not self.send_numlist(L):
			self.log.warning("Query aborted because send failed.")
			return False
		
		# Receive encrypted message and save to 'reply'
		self.reply = self.recv_str()
		if self.reply is None:
			self.log.error(f"query returned None. (msg: {L})")
			return False
		
		# self.log.debug(f"QUERY(): Received '{self.reply}'")
		
		# Populate reply_ec if error detected
		if self.reply[0:6] == "ERROR:":
			self.error_code = self.reply[6:]

		return True
	
	def exit(self):
		""" Tells the server to drop the connection """
		
		# Check for valid state
		if self.state != ClientAgent.CS_LOGIN:
			self.log.warning("Cannot exit except in LOGIN state.")
			return False
		
		self.log.debug("Exit Process Beginning")

		# # Set online status to false
		# self.online = False
		
		if not self.query("EXIT"):
			self.log.warning("Failed to send EXIT signal")
			return False
		
		if self.reply != "PASS":
			self.log.warning("Server failed to process exit!")
			return False
		
		self.log.info("Successfully exited server.")
		self.state = ClientAgent.CS_LOGIN
		# self.online = False
				
		return True
	
	def logout(self):
		""" Logs the user out, deauthorizing the client to act on their behalf."""
		
		# Check for valid state
		if self.state != ClientAgent.CS_MAIN:
			self.log.warning(f"Cannot logout except in MAIN state. (state = {self.state})")
			return False
		
		if not self.query("LOGOUT"):
			self.log.warning("Failed to send LOGOUT signal")
			return False
		
		if self.reply != "PASS":
			self.log.warning("Server failed to process logout!")
			return False
		
		self.log.info("Successfully logged off server.")
		self.state = ClientAgent.CS_LOGIN
		# self.online = False
		
		self.user = None
		
		return True
				
	def login(self, username:str, password:str):
		""" Execute login sequence.

		It will return true or false depending on if login was successful.
		"""
		
		self.user = None
		
		# Check for valid state
		if self.state != ClientAgent.CS_LOGIN:
			self.log.warning("Cannot login except in LOGIN state.")
			return False
		
		self.log.debug("Login Process Beginning")
		
		# Begin Login sequence, expect request for username and password
		if (not self.query("LOGIN")) or (self.reply != "UNPWD"):
			self.log.debug("Query check for LOGIN failed.")
			
			# self.online = False
			return False

		# Send username and password (encrypted, but not hashed)
		
		# Send username
		if not self.query(username) or self.reply != "ACK":
			
			# self.online = False
			return False
		
		# Send username
		if not self.query(password) or self.reply != "ACK":
			
			# self.online = False
			return False
		
		# Get login status
		if not self.query("STATUS?"):
			
			# self.online = False
			return False
		
		if self.reply == "PASS":
			# self.online = True
			
			self.user = username
			self.state = ClientAgent.CS_MAIN
			
			return True
		else:
			# self.online = False
			return False
	
	def create_account(self, username:str, email:str, password:str):
		""" Creates a new account on the server. """
		
		# Check for valid state
		if self.state != ClientAgent.CS_LOGIN:
			self.log.warning("Cannot create an account except in LOGIN state.")
			return False
		
		# Begin Signup sequence, expect request for email
		if not self.query("SIGNUP") or self.reply != "EMAIL":
			
			# self.online = False
			return False
		
		# Send email username and password (encrypted, but not hashed)
		
		# Send email
		if not self.query(email) or self.reply != "UNPWD":
			
			# self.online = False
			return False
		
		# Send username
		if not self.query(username) or self.reply != "ACK":
			
			# self.online = False
			return False
		
		# Send password
		if not self.query(password) or self.reply != "ACK":
			
			# self.online = False
			return False
		
		# Get login status
		if not self.query("STATUS?"):
			
			# self.online = False
			return False
		
		if self.reply == "PASS":
			# self.online = True
			return True
		else:
			# self.online = False
			return False
		
	def handshake(self):
		""" Performs handshake with server, exchanging keys. Remainder
		of traffic will be encrypted. """
		
		# Check for valid state
		if self.state != ClientAgent.CS_HAND:
			self.log.warning(f"Cannot perform handshake except in CD_HAND state. (state = {self.state})")
			return False
		
		self.log.debug("Initiating handshake")

		# # Set online status to false
		# self.online = False

		# Tell server to begin handshake
		send_ptstring(self.sock, "HS")

		# Get server private key back
		self.log.debug("Receiving public key")
		self.server_key = rsa.PublicKey.load_pkcs1(self.sock.recv(1024))

		# Send client private key
		self.log.debug("Sending public key")
		self.sock.send(self.public_key.save_pkcs1("PEM"))

		# self.online = True

		# Receive sync packet
		cmd = self.rsa_recv() # This should be void, without this, the sockets will freeze
		
		## At this point, the handshake is complete and comms are encrypted with RSA ####
		
		# Get AES key
		self.rsa_send("AES_KEY")
		self.log.debug("Waiting for RSA key")
		self.aes_key = self.rsa_recv()
		
		# Get AES iv
		self.rsa_send("AES_IV")
		self.log.debug("Waiting for RSA iv")
		self.aes_iv = self.rsa_recv()

		### At this point the AES keys have been exchanged and the handshake is complete
		
		# Update state
		self.state = ClientAgent.CS_LOGIN
		
		self.log.info("Completed handshake")
		
	def view_database(self):
		""" Retrieve the database as a string. Admin only"""
		
		# Check for valid state
		if self.state != ClientAgent.CS_MAIN:
			self.log.warning("Cannot view database except in MAIN state.")
			return None
		
		# query database
		if not self.query("VIEWDB"):
			
			return None
		else:
			return self.reply
	
	def shutdown_server(self):
		""" Retrieve the database as a string. Admin only"""
		
		# Check for valid state
		if self.state != ClientAgent.CS_MAIN:
			self.log.warning("Cannot shutdown server except in MAIN state.")
			return False
		
		# query database
		if not self.query("SHUTDOWN"):
			
			return False
		
		# Return result
		return (self.reply == "PASS")
	
	def delete_account(self, username:str):
		""" Delete the user account for 'username'. Admin only"""
		
		# Check for valid state
		if self.state != ClientAgent.CS_MAIN:
			self.log.warning("Cannot delete accounts except in MAIN state.")
			return False
		
		# Check username is valid before transmitting to server
		if not username_follows_rules(username):
			self.log.warning("Invalid username provided.")
			return False
		
		# query database
		if not self.query("DELUSR"):
			
			return False
		
		# Return result
		if self.reply != "USR":
			
			return False
		
		# query database
		if not self.query(username):
			
			return False
		
		# Return result - Note this will return as 'PASS' even if the user didn't exist
		return (self.reply == "PASS")
	
	def message_user(self, user:str, message:str):
		
		""" Sends a message to the user (specified by username) """
		
		# Check for valid state
		if self.state == ClientAgent.CS_HAND or self.state == ClientAgent.CS_LOGIN:
			self.log.warning("Cannot message users prior to login.")
			return False
		
		# Check username is valid before transmitting to server
		if not username_follows_rules(user):
			self.log.warning("Invalid username provided.")
			return False
		
		# Verify message doesn't contain unrecognized characters
		filt_msg = validate_message(message)
		
		# Send message
		if not self.query("MSGUSR"):
			self.log.warning(f"Failed to send MSGUSR signal ({self.err()})")
			return False
		
		if self.reply != "USR":
			self.log.warning(f"Server failed to process MSGUSR ({self.err()})")
			return False
		
		# Send recipient name
		if not self.query(user):
			self.log.warning(f"Failed to send MSGUSR signal ({self.err()})")
			return False
		
		if self.reply != "MSG":
			self.log.warning(f"Server failed to process MSGUSR ({self.err()})")
			return False
		
		# Send message
		if not self.query(filt_msg):
			self.log.warning(f"Failed to send MSGUSR signal ({self.err()})")
			return False
		
		if self.reply != "PASS":
			self.log.warning(f"Server failed to process message in MSGUSR exchange ({self.err()})")
			return False
	
	def sync(self):
		""" Updates local data from the server. Gets messages/notifications, game data, etc """
		
		# Check for valid state
		if self.state == ClientAgent.CS_HAND or self.state == ClientAgent.CS_LOGIN:
			self.log.warning("Cannot sync prior to login.")
			return False
		
		# query database
		if not self.query("SYNC"):
			return False
		
		# Get returned result
		try:
			sd = SyncData()
			sd.from_utf8(self.reply.encode()) #TODO: This is repetative! query automatically decodes, but from_utf8 wants it encoded again!
		except Exception as e:
			self.log.error(f"Failed to read SyncData ({e})")
			self.log.debug(f"SyncData: {self.reply}")
			return
		
		###### NOw we have the JSON data saved in JD. Turn it into a SyncData object #######
		
		# Transfer data from SyncData object to ClientAgent object
		try:
			self.notes = sd.notes
			self.game.unpack(sd.packed_game)
		except Exception as e:
			self.log.warning(f"Failed to populate ClientAgent from SyncData ({e})")
			return
		
		if self.game.id == -1:
			self.state = ClientAgent.CS_MAIN
		
		elif self.game.state == Game.GS_RUN:
			self.state = ClientAgent.CS_GAME
			
		elif self.game.state == Game.GS_LOBBY:
			self.state = ClientAgent.CS_LOBBY
			
		elif self.game.state == Game.GS_END:
			self.state = ClientAgent.CS_RESULTS


from heimdallr.base import *

class VectorNetworkAnalyzerCtg(Driver):
	
	MEAS_S11 = "meas-s11"
	MEAS_S21 = "meas-s21"
	MEAS_S12 = "meas-s12"
	MEAS_S22 = "meas-s22"
	
	def __init__(self, address:str, log:LogPile):
		super().__init__(address, log)
	
	def set_freq_start(self, f_Hz:float, channel:int=1):
		pass
	
	def set_freq_end(self, f_Hz:float, channel:int=1):
		pass
	
	def set_power(self, p_dBm:float, channel:int=1):
		pass
	
	def set_num_points(self, points:int, channel:int=1):
		pass
	
	def set_res_bandwidth(self, rbw_Hz:float, channel:int=1):
		pass
	
	def clear_traces(self):
		pass
	
	def add_trace(self, channel:int, measurement:str):
		''' Returns trace number '''
		pass
	
	def get_trace(self, trace:int):
		
		pass


from heimdallr.base import *

class RFSignalGeneratorCtg(Driver):
	
	def __init__(self, address:str, log:LogPile, expected_idn:str=""):
		super().__init__(address, log, expected_idn=expected_idn)
	
	def set_power(self, p_dBm:float):
		pass
	
	def set_freq(self, f_Hz:float):
		pass
	
	def set_enable_rf(self,enable:bool):
		pass


from heimdallr.base import *

class PIDTemperatureControllerCtg(Driver):
	
	def __init__(self, address:str, log:LogPile):
		super().__init__(address, log)
	
	def set_temp(self, temp_K:float, channel:int=1):
		pass
	
	def get_temp(self, temp_K:float, channel:int=1):
		pass
	
	def set_enable(self, enable:bool, channel:int=1):
		pass


''' Driver for PNA E8364B VNA

* Only supports a single window

Manual: https://www.testworld.com/wp-content/uploads/user-guide-help-agilent-e8362b-e8363b-e8364b-e8361a-n5230a-n5242a-pna-series-microwave-network-analyzers.pdf
'''

from heimdallr.base import *
from heimdallr.instrument_control.categories.vector_network_analyzer_ctg import *

class KeysightPNAE8364B(VectorNetworkAnalyzerCtg):
	
	SWEEP_CONTINUOUS = "sweep-continuous"
	SWEEP_SINGLE = "sweep-single"
	SWEEP_OFF = "sweep-off"
	
	def __init__(self, address:str, log:LogPile):
		super().__init__(address, log)
		
		self.trace_lookup = {}
		
		self.measurement_codes = {}
		self.measurement_codes[VectorNetworkAnalyzerCtg.MEAS_S11] = "S11"
		self.measurement_codes[VectorNetworkAnalyzerCtg.MEAS_S12] = "S12"
		self.measurement_codes[VectorNetworkAnalyzerCtg.MEAS_S21] = "S21"
		self.measurement_codes[VectorNetworkAnalyzerCtg.MEAS_S22] = "S22"
	
	def set_freq_start(self, f_Hz:float, channel:int=1):
		self.write(f"SENS{channel}:FREQ:STAR {f_Hz}")
	def get_freq_start(self, channel:int=1):
		return float(self.query(f"SENS{channel}:FREQ:STAR?"))
	
	def set_freq_end(self, f_Hz:float, channel:int=1):
		self.write(f"SENS{channel}:FREQ:STOP {f_Hz}")
	def get_freq_end(self, channel:int=1):
		return float(self.query(f"SENS{channel}:FREQ:STOP?"))
	
	def set_power(self, p_dBm:float, channel:int=1, port:int=1):
		self.write(f"SOUR{channel}:POW{port}:LEV:IMM:AMPL {p_dBm}")
	def get_power(self, channel:int=1, port:int=1):
		return float(self.query(f"SOUR{channel}:POW{port}:LEV:IMM:AMPL?"))
	
	def set_num_points(self, points:int, channel:int=1):
		self.write(f"SENS{channel}:SWEEP:POIN {points}")
	def get_num_points(self, channel:int=1):
		return int(self.query(f"SENS{channel}:SWEEP:POIN?"))
	
	def set_res_bandwidth(self, rbw_Hz:float, channel:int=1):
		self.write(f"SENS{channel}:BAND:RES {rbw_Hz}")
	def get_res_bandwidth(self, channel:int=1):
		return float(self.query(f"SENS{channel}:BAND:RES?"))
	
	def set_rf_enable(self, enable:bool):
		self.write(f"OUTP:STAT {bool_to_ONFOFF(enable)}")
	def get_rf_enable(self):
		return str_to_bool(self.query(f"OUTP:STAT?"))
	
	def clear_traces(self):
		self.write(f"CALC:PAR:DEL:ALL")
	
	def add_trace(self, channel:int, trace:int, measurement:str):
		
		# Get measurement code
		try:
			meas_code = self.measurement_codes[measurement]
		except:
			self.log.error(f"Unrecognized measurement!")
			return
		
		# Check that trace doesn't already exist
		if trace in self.trace_lookup.keys():
			self.log.error(f"Cannot add trace. Trace number {trace} already exists.")
		
		# Create name and save
		trace_name = f"trace{trace}"
		self.trace_lookup[trace] = trace_name
		
		# Create measurement - will not display yet
		self.write(f"CALC{channel}:PAR:DEF '{trace_name}', {meas_code}")
		
		# Create a trace and assoc. with measurement
		self.write(f"DISP:WIND:TRAC{trace}:FEED '{trace_name}'")
	
	def get_trace_data(self, channel:int, trace:int):
		
		# Check that trace exists
		if trace not in self.trace_lookup.keys():
			self.log.error(f"Trace number {trace} does not exist!")
			return
		
		trace_name = self.trace_lookup[trace]
		
		# Select the specified measurement/trace
		self.write(f"CALC{channel}:PAR:SEL {trace_name}")
		
		# Set data format
		self.write(f"FORM:DATA REAL,64")
		
		# Query data
		return self.query(f"CALC{channel}:DATA? SDATA")
		
	def set_continuous_trigger(self, enable:bool):
		self.write(f"INIT:CONT {bool_to_ONFOFF(enable)}")
	def get_continuous_trigger(self):
		return str_to_bool(self.query(f"INIT:CONT?"))
	
	def send_manual_trigger(self):
		self.write(f"INIT:IMM")
		
	def set_averaging_enable(self, enable:bool, channel:int=1):
		self.write(f"SENS{channel}:AVER {bool_to_ONFOFF(enable)}")
	def get_averaging_enable(self, channel:int=1):
		return str_to_bool(self.write(f"SENS{channel}:AVER?"))
	
	def set_averaging_count(self, count:int, channel:int=1):
		count = int(max(1, min(count, 65536)))
		if count != count:
			self.log.error(f"Did not apply command. Instrument limits values to integers 1-65536 and this range was violated.")
			return
		self.write(f"SENS{channel}:AVER:COUN {count}")
	def get_averaging_count(self, channel:int=1):
		return int(self.query(f"SENS{channel}:AVER:COUN?"))
	
	def send_clear_averaging(self, channel:int=1):
		self.write(f"SENS{channel}:AVER:CLE")
	
	def send_preset(self):
		self.write("SYST:PRES")


""" Keysight 8360L Series Swept CW Generator

Manual: http://www.doe.carleton.ca/~nagui/labequip/synth/manuals/e4400324.pdf
"""

from heimdallr.base import *
from heimdallr.instrument_control.categories.rf_signal_generator_ctg import *

class AgilentE4400(RFSignalGeneratorCtg):

	def __init__(self, address:str, log:LogPile):
		super().__init__(address, log, expected_idn='Hewlett-Packard, ESG-4000B')
	
	def set_power(self, p_dBm:float):
		self.write(f":POW:LEV:IMM:AMPL {p_dBm} dBm")
	def get_power(self):
		val = self.query(f":POW:LEV:IMM:AMPL?")
		return float(val)
	
	def set_freq(self, f_Hz:float):
		self.write(f":FREQ:CW {f_Hz} Hz")
	def get_freq(self):
		return float(self.query(f":FREQ:CW?"))
	
	def set_enable_rf(self, enable:bool):
		self.write(f":OUTP:STAT {bool_to_str01(enable)}")
	def get_enable_rf(self):
		return str_to_bool(self.query(f":OUTP:STAT?"))


''' Driver for Rohde & Schwarz NRX Power Meter 

Manual: https://scdn.rohde-schwarz.com/ur/pws/dl_downloads/pdm/cl_manuals/user_manual/1178_5566_01/NRX_UserManual_en_10.pdf
'''

from heimdallr.base import *
from heimdallr.instrument_control.categories.rf_power_sensor_ctg import *

class RohdeSchwarzNRX(RFPowerSensor):
	
	def __init__(self, address:str, log:LogPile):
		super().__init__(address, log, expected_idn="Rohde&Schwarz,NRX") # Example string:  'Rohde&Schwarz,NRX,1424.7005k02/102854,02.40.20100501\n'
		
	def set_meas_frequency(self, f_Hz:float):
		self.write(f"SENSE:FREQ:CW {f_Hz}")
	def get_meas_frequency(self):
		return float(self.query(f"SENSE:FREQ:CW?"))
	
	# def send_autoscale(self):
	# 	print("Should this be send or enable? Is it one time?")
	# 	self.write(f":SENS:POW:RANGE:AUTO ON")
	
	def send_trigger(self, wait:bool=False):
		
		self.write(f"*CLS") # Clear status register so can determine when ready
		
		# Maybe should be :INIT:ALL? I think this just triggers all sensors vs one
		# Send trigger
		self.write(f":INIT:IMM")
		
		# Wait for operation to complete if requested
		if wait:
			self.wait_ready()
		
	
	def get_measurement(self):
		data = float(self.query(f"CALC1:DATA?"))
		return data
	
	def set_averaging_count(self, counts:int, meas_no:int=1):
		
		 # Enforce bounds - counts
		counts = max(1, min(counts, 1048576))
		if counts != counts:
			self.log.error(f"Did not apply command. Instrument limits number of counts from 1 to 1048576 and this range was violated.")
			return
		
		# Enforce bounds - meas_no
		meas_no = max(1, min(meas_no, 8))
		if meas_no != meas_no:
			self.log.error(f"Did not apply command. Instrument limits measurement-number values from 1 to 8 and this range was violated.")
			return
		
		# Legacy version, works for R&S but deprecated
		#  [SENSe<Sensor>:]AVERage:COUNt[:VALue]
		
		self.write(f"CALC{meas_no}:CHAN1:AVER:COUN:VAL {counts}")
		
	# def send_trigger(self):
	# 	self.write(f"*CLS")
	# 	self.write(f"INIT:ALL")
	# 	wait(nrx)
	
	# def get_measurement(self):
		
	# 	# FRom the documentation:
	# 	#   The FETCh? and READ? commands send without parameters have a special meaning.
	# 	#   FETCh? returns the current measured value if it is valid. If a measured value is not yet
	# 	#    available, processing is suspended until a valid result is available.
	# 	return float(self.query(f"FETCH?"))



""" LakeShore Model 335 Temperature Controller Driver


Manual: https://www.lakeshore.com/docs/default-source/product-downloads/335_manual038a7cfe0db7421f941ebb45db85741f.pdf?sfvrsn=e16b9529_1
"""

from heimdallr.instrument_control.categories.all_ctgs import *

class LakeShoreModel335(PIDTemperatureControllerCtg):
	
	RANGE_OFF = 0
	RANGE_LOW = 1
	RANGE_MID = 2
	RANGE_HIGH = 3
	
	def __init__(self, address:str, log:LogPile):
		super().__init__(address, log)
	
	def set_setpoint(self, temp_K:float, channel:int=1):
		self.inst.write(f"SETP {channel},{temp_K}")
	def get_setpoint(self, channel:int=1):
		return float(self.inst.query(f"SETP? {channel}"))
	
	def get_temp(self, temp_K:float, channel:int=1):
		return float(self.inst.query(f"SETP? {channel}"))
	
	def set_pid(self, P:float, I:float, D:float, channel:int=1):
		P = max(0.1, min(P, 1000))
		I = max(0.1, min(I, 1000))
		D = max(0.1, min(D, 1000))
		
		# Print a warning if any variable was adjusted
		if P != P or I != I or D != D:
			self.log.error(f"Did not apply command. Instrument limits values to 0.1-1000 and this range was violated.")
			return
		self.write(f"PID {channel}")
	
	def get_pid(self, channel:int=1):
		
		pid_str = self.query(f"PID? {channel}")
		print(pid_str)
		
		# Split at commas
		pidstr_list = pid_str.split(",")
		
		# Format into array of numbers
		pid_list = [float(x) for x in pidstr_list]
		
		# Return values as dictionary
		return {"P": pid_list[0], "I": pid_list[1], "D": pid_list[2]}
	
	def set_range(self, range:int, channel:int=1):
		
		range = max(0, min(range, 3))
		if range != range:
			self.log.error(f"Did not apply command. Invalid range parameter supplied. Use range constants built into class.")
			return
		
		self.write(f"RANGE {channel},{range}")
	def get_range(self, channel:int=1):
		return int(self.query(f"RANGE? {channel}"))
	
	def set_enable(self, enable:bool, channel:int=1):
		pass
		#PRobably OUTMODE command, but should play around with this


''' Driver for Rohde & Schwarz NRX Power Meter 

Manual: https://scdn.rohde-schwarz.com/ur/pws/dl_downloads/pdm/cl_manuals/user_manual/1178_5566_01/NRX_UserManual_en_10.pdf
	    https://scdn.rohde-schwarz.com/ur/pws/dl_downloads/pdm/cl_manuals/user_manual/1177_5079_01/NRPxxSN_UserManual_en_21.pdf
'''

from heimdallr.base import *
from heimdallr.instrument_control.categories.rf_power_sensor_ctg import *

class RohdeSchwarzNRP(RFPowerSensor):
	
	def __init__(self, address:str, log:LogPile):
		super().__init__(address, log, expected_idn="Rohde&Schwarz,NRP") # Example string:  ''
		
	def set_meas_frequency(self, f_Hz:float):
		self.write(f"SENSE:FREQUENCY {f_Hz}")
	def get_meas_frequency(self):
		return float(self.query(f"SENSE:FREQUENCY?"))
	
	# def send_autoscale(self):
	# 	print("Should this be send or enable? Is it one time?")
	# 	self.write(f":SENS:POW:RANGE:AUTO ON")
	
	def send_trigger(self, wait:bool=False):
		
		# Set unit to dBm - if done after trigger, seems to void results
		self.write("UNIT:POW DBM")
		
		self.write(f"*CLS") # Clear status register so can determine when ready
		
		self.write("INITIATE")
		
		# Wait for operation to complete if requested
		if wait:
			self.wait_ready()
		
	
	def get_measurement(self):
		
		data = float(self.query(f"FETCH?"))
		return data
	
	# def set_averaging_count(self, counts:int, meas_no:int=1):
		
	# 	 # Enforce bounds - counts
	# 	counts = max(1, min(counts, 1048576))
	# 	if counts != counts:
	# 		self.log.error(f"Did not apply command. Instrument limits number of counts from 1 to 1048576 and this range was violated.")
	# 		return
		
	# 	# Enforce bounds - meas_no
	# 	meas_no = max(1, min(meas_no, 8))
	# 	if meas_no != meas_no:
	# 		self.log.error(f"Did not apply command. Instrument limits measurement-number values from 1 to 8 and this range was violated.")
	# 		return
		
	# 	# Legacy version, works for R&S but deprecated
	# 	#  [SENSe<Sensor>:]AVERage:COUNt[:VALue]
		
	#	self.write(f"CALC{meas_no}:CHAN1:AVER:COUN:VAL {counts}")




from heimdallr.instrument_control.categories.oscilloscope_ctg import *
from heimdallr.instrument_control.categories.vector_network_analyzer_ctg import *
from heimdallr.instrument_control.categories.temperature_controller_ctg import *
from heimdallr.instrument_control.categories.rf_signal_generator_ctg import *
from heimdallr.instrument_control.categories.spectrum_analyzer_ctg import *
from heimdallr.instrument_control.categories.rf_power_sensor_ctg import *


''' Driver for Rohde & Schwarz FSQ series Spectrum Analyzers

* Only supports a single window (Referred to as Screen A in R&S documentation. The instruments supports screens A&B.)

Manual: https://scdn.rohde-schwarz.com/ur/pws/dl_downloads/dl_common_library/dl_manuals/gb_1/f/fsq_1/FSQ_OperatingManual_en_02.pdf
'''

import array
from heimdallr.base import *
from heimdallr.instrument_control.categories.spectrum_analyzer_ctg import *

class RohdeSchwarzFSQ(SpectrumAnalyzerCtg):
	
	def __init__(self, address:str, log:LogPile):
		super().__init__(address, log, expected_idn="Rohde&Schwarz,FSQ-") # Example 'Rohde&Schwarz,FSQ-26,200334/026,4.75\n'
		
		self.trace_lookup = {}
	
	def set_freq_start(self, f_Hz:float):
		self.write(f"SENS:FREQ:STAR {f_Hz} Hz")
	def get_freq_start(self):
		return float(self.query(f"SENS:FREQ:STAR?"))
	
	def set_freq_end(self, f_Hz:float):
		self.write(f"SENS:FREQ:STOP {f_Hz}")
	def get_freq_end(self,):
		return float(self.query(f"SENS:FREQ:STOP?"))
	
	def set_ref_level(self, ref_dBm:float):
		ref_dBm = max(-130, min(ref_dBm, 30))
		if ref_dBm != ref_dBm:
			self.log.error(f"Did not apply command. Instrument limits values from -130 to 30 dBm and this range was violated.")
			return
		self.write(f"CALC:UNIT:POW dBm") # Set units to DBM (Next command refers to this unit)
		self.write(f"DISP:WIND:TRAC:Y:RLEV {ref_dBm}")
	def get_ref_level(self):
		return float(self.query("DISP:WIND:TRAC:Y:RLEV?"))
	
	def set_y_div(self, step_dB:float):
		
		step_dB = max(1, min(step_dB, 20))
		if step_dB != step_dB:
			self.log.error(f"Did not apply command. Instrument limits values from 1 to 20 dB and this range was violated.")
			return
		
		full_span_dB = step_dB*10 #Sets total span, not per div, so must multiply by num. divisions (10)
		self.write(f":DISP:WIND:TRAC:Y:SCAL {full_span_dB} DB") 
	def get_y_div(self):
		full_span_dB = float(self.query(f":DISP:WIND:TRAC:Y:SCAL?"))
		return full_span_dB/10
	
	def set_res_bandwidth(self, rbw_Hz:float, channel:int=1):
		self.write(f"SENS:BAND:RES {rbw_Hz} Hz")
	def get_res_bandwidth(self, channel:int=1):
		return float(self.query(f"SENS:BAND:RES?"))
	
	def set_continuous_trigger(self, enable:bool):
		self.write(f"INIT:CONT {bool_to_ONFOFF(enable)}")
	def get_continuous_trigger(self):
		return str_to_bool(self.query(f"INIT:CONT?"))
	
	def send_manual_trigger(self, send_cls:bool=True):
		if send_cls:
			self.write("*CLS")
		self.write(f"INIT:IMM")
	
	def get_trace_data(self, trace:int, use_ascii_transfer:bool=False, use_fast_binary:bool=True):
		''' Returns the data of the trace in a standard waveform dict, which
		
		has keys:
			* x: X data list (float)
			* y: Y data list (float)
			* x_units: Units of x-axis
			* y_units: UNits of y-axis
		
		'''
		
		# Make sure trace is in range
		count = int(max(1, min(trace, 3)))
		if count != count:
			self.log.error(f"Did not apply command. Instrument limits values to integers 1-3 and this range was violated.")
			return
		
		# Get Y-unit
		
		
		# Run ASCII transfer if requested
		if use_ascii_transfer:
			
			self.write(f"FORMAT:DATA ASCII") # Set format to ASCII
			data_raw = self.query(f"TRACE:DATA? TRACE{trace}") # Get raw data
			str_list = data_raw.split(",") # Split at each comma
			del str_list[-1] # Remove last element (newline)
			float_data = [float(x) for x in str_list] # Convert to float
			
		else:
			#  Example data would be:
			#      #42500<data block of 2500 4 byte floats>
			#	   THe '#4' indicates 4 bytes of data for size of packet
			#      The 2500 indicates 2500 floats, or 2500*4 bytes
		
			# Set data format - Real 32 binary data - in current Y unit
			self.write(f"FORMAT:DATA REAL")
			
			if not use_fast_binary:
			
				# Read data - ask for data
				self.write(f"TRACE:DATA? TRACE{trace}")
				data_raw = bytearray()
				while True:
					try:
						byte = self.inst.read_bytes(1)
						data_raw += byte
					except:
						break
				
				# Skip first 6 bytes (number of elements) and last byte (newline)
				float_data = list(array.array('f', data_raw[6:-1]))
				
			else:
				
				# Read data - ask for data
				self.write(f"TRACE:DATA? TRACE{trace}")
				data_raw = bytearray()
				
				# Get size of size of packet block (ie. convert #4 -> (int)4 )
				byte = self.inst.read_bytes(2)
				data_raw += byte
				digits_in_size_num = int(data_raw[1:2])
				
				# Read size of packet
				byte = self.inst.read_bytes(digits_in_size_num)
				data_raw += byte
				packet_size = int(data_raw[2:2+digits_in_size_num])
				
				#Read entire packet
				byte = self.inst.read_bytes(packet_size+1)
				data_raw += byte
				
				# Skip first X bytes (number of elements) and last byte (newline)
				float_data = list(array.array('f', data_raw[2+digits_in_size_num:-1]))
				
		# Generate time array
		f_list = list(np.linspace(self.get_freq_start(), self.get_freq_end(), len(float_data)))
		
		out_data = {'x':f_list, 'y':float_data, 'x_units':'Hz', 'y_units':'dBm'}
		
		# Convert Y-unit to dBm
		return out_data
		
		# trace_name = self.trace_lookup[trace]
		
		# # Select the specified measurement/trace
		# self.write(f"CALC{channel}:PAR:SEL {trace_name}")
		
		# # Set data format
		# self.write(f"FORM:DATA REAL,64")
		
		# # Query data
		# return self.query(f"CALC{channel}:DATA? SDATA")
		
	
		
	# def set_averaging_enable(self, enable:bool, channel:int=1):
	# 	self.write(f"SENS{channel}:AVER {bool_to_ONFOFF(enable)}")
	# def get_averaging_enable(self, channel:int=1):
	# 	return str_to_bool(self.write(f"SENS{channel}:AVER?"))
	
	# def set_averaging_count(self, count:int, channel:int=1):
	# 	count = int(max(1, min(count, 65536)))
	# 	if count != count:
	# 		self.log.error(f"Did not apply command. Instrument limits values to integers 1-65536 and this range was violated.")
	# 		return
	# 	self.write(f"SENS{channel}:AVER:COUN {count}")
	# def get_averaging_count(self, channel:int=1):
	# 	return int(self.query(f"SENS{channel}:AVER:COUN?"))
	
	# def send_clear_averaging(self, channel:int=1):
	# 	self.write(f"SENS{channel}:AVER:CLE")
	
	# def send_preset(self):
	# 	self.write("SYST:PRES")


""" Keysight 8360L Series Swept CW Generator
"""

from heimdallr.base import *
from heimdallr.instrument_control.categories.rf_signal_generator_ctg import *

class Keysight8360L(RFSignalGeneratorCtg):

	def __init__(self, address:str, log:LogPile):
		# Example: "HEWLETT-PACKARD,83650L,3844A00476,19 JAN 00\n"
		super().__init__(address, log, expected_idn="HEWLETT-PACKARD,836")
		
	
	def set_power(self, p_dBm:float):
		self.write(f":POW:LEV {p_dBm}")
	def get_power(self):
		val = self.query(f":POW:LEV?")
		return float(val)
	
	def set_freq(self, f_Hz:float):
		self.inst.write(f":SOUR:FREQ:CW {f_Hz}")
	def get_freq(self):
		return float(self.inst.query(f":SOUR:FREQ:CW?"))
	
	def set_enable_rf(self, enable:bool):
		self.inst.write(f":OUTP:STAT {bool_to_str01(enable)}")
	def get_enable_rf(self):
		return str_to_bool(self.inst.query(f":OUTP:STAT?"))


"""RIGOL’s 1000Z Series Digital Oscilloscope

https://beyondmeasure.rigoltech.com/acton/attachment/1579/f-0386/1/-/-/-/-/DS1000Z_Programming%20Guide_EN.pdf
"""

from heimdallr.instrument_control.categories.all_ctgs import *

class RigolDS1000Z(Oscilloscope2Ctg):

	def __init__(self, address:str, log:LogPile):
		super().__init__(address, log, expected_idn='RIGOL TECHNOLOGIES,DS10')
		
		self.meas_table = {Oscilloscope2Ctg.MEAS_VMAX:'VMAX', Oscilloscope2Ctg.MEAS_VMIN:'VMIN', Oscilloscope2Ctg.MEAS_VAVG:'VAVG', Oscilloscope2Ctg.MEAS_VPP:'VPP', Oscilloscope2Ctg.MEAS_FREQ:'FREQ'}
		
		self.stat_table = {Oscilloscope2Ctg.STAT_AVG:'AVER', Oscilloscope2Ctg.STAT_MAX:'MAX', Oscilloscope2Ctg.STAT_MIN:'MIN', Oscilloscope2Ctg.STAT_CURR:'CURR', Oscilloscope2Ctg.STAT_STD:'DEV'}
		
	def set_div_time(self, time_s:float):
		self.write(f":TIM:MAIN:SCAL {time_s}")
	def get_div_time(self):
		return self.query(f":TIM:MAIN:SCAL?")
	
	def set_offset_time(self, channel:int, time_s:float):
		self.write(f":TIM:MAIN:OFFS {time_s}")
	def get_offset_time(self, channel:int, time_s:float):
		return self.query(f":TIM:MAIN:OFFS?")
	
	def set_div_volt(self, channel:int, volt_V:float):
		self.write(f":CHAN{channel}:SCAL {volt_V}")
	def get_div_volt(self, channel:int, volt_V:float):
		return self.query(f":CHAN{channel}:SCAL?")
	
	def set_offset_volt(self, channel:int, volt_V:float):
		self.write(f":CHAN{channel}:OFFS {volt_V}")
	def get_offset_volt(self, channel:int, volt_V:float):
		return self.query(f":CHAN{channel}:OFFS?")
	
	def set_chan_enable(self, channel:int, enable:bool):
		self.write(f":CHAN{channel}:DISP {bool_to_str01(enable)}")
	def get_chan_enable(self, channel:int):
		val_str = self.query(f":CHAN{channel}:DISP?")
		return str01_to_bool(val_str)
	
	def get_waveform(self, channel:int):
		
		self.write(f"WAV:SOUR CHAN{channel}")  # Specify channel to read
		self.write("WAV:MODE NORM")  # Specify to read data displayed on screen
		self.write("WAV:FORM ASCII")  # Specify data format to ASCII
		data = self.query("WAV:DATA?")  # Request data
		
		if data is None:
			return {"time_s":[], "volt_V":[]}
		
		# Split string into ASCII voltage values
		volts = data[11:].split(",")
		
		volts = [float(v) for v in volts]
		
		# Get timing data
		xorigin = float(self.query("WAV:XOR?"))
		xincr = float(self.query("WAV:XINC?"))
		
		# Get time values
		t = list(xorigin + np.linspace(0, xincr * (len(volts) - 1), len(volts)))
		
		return {"time_s":t, "volt_V":volts}
	
	def add_measurement(self, meas_type:int, channel:int=1):
		
		# Find measurement string
		if meas_type not in self.meas_table:
			self.log.error(f"Cannot add measurement >{meas_type}<. Measurement not recognized.")
			return
		item_str = self.meas_table[meas_type]
		
		# Get channel string
		channel = max(1, min(channel, 1000))
		if channel != channel:
			self.log.error("Channel must be between 1 and 4.")
			return
		src_str = f"CHAN{channel}"
		
		# Send message
		self.write(f":MEASURE:ITEM {item_str},{src_str}")
	
	def get_measurement(self, meas_type:int, channel:int=1, stat_mode:int=0) -> float:
		
		# FInd measurement string
		if meas_type not in self.meas_table:
			self.log.error(f"Cannot add measurement >{meas_type}<. Measurement not recognized.")
			return
		item_str = self.meas_table[meas_type]
		
		# Get channel string
		channel = max(1, min(channel, 1000))
		if channel != channel:
			self.log.error("Channel must be between 1 and 4.")
			return
		src_str = f"CHAN{channel}"
		
		
		
		# Query result
		if stat_mode == 0:
			return self.query(f":MEASURE:ITEM? {item_str},{src_str}")
		else:
			
			# Get stat string
			if stat_mode not in self.stat_table:
				self.log.error(f"Cannot use statistic option >{meas_type}<. Option not recognized.")
				return
			stat_str = self.stat_table[stat_mode]
			
			return self.query(f":MEASURE:STAT:ITEM? {stat_str},{item_str},{src_str}")
	
	def clear_measurements(self):
		
		self.write(f":MEASURE:CLEAR ALL")
	
	def set_measurement_stat_display(self, enable:bool):
		'''
		Turns display statistical values on/off for the Rigol DS1000Z series scopes. Not
		part of the OscilloscopeCtg, but local to this driver.
		
		Args:
			enable (bool): Turns displayed stats on/off
		
		Returns:
			None
		'''
		
		self.write(f":MEASure:STATistic:DISPlay {bool_to_ONFOFF(enable)}")


from heimdallr.instrument_control.drivers.PNA_dvr import *
from heimdallr.instrument_control.drivers.Rigol_DS1000Z_dvr import *
from heimdallr.instrument_control.drivers.Keysight_8360L_dvr import *
from heimdallr.instrument_control.drivers.LakeShore_Model335_dvr import *
from heimdallr.instrument_control.drivers.Siglent_SSA3000X_dvr import *
from heimdallr.instrument_control.drivers.Agilent_E4400_dvr import *
from heimdallr.instrument_control.drivers.RohdeSchwarz_NRX_dvr import *
from heimdallr.instrument_control.drivers.RohdeSchwarz_NRP_dvr import *
from heimdallr.instrument_control.drivers.RohdeSchwarz_FSQ_dvr import *


''' Driver for Siglent SSA3000X Spectrum Analyzers

* Only supports a single window

Manual: https://www.testworld.com/wp-content/uploads/user-guide-help-agilent-e8362b-e8363b-e8364b-e8361a-n5230a-n5242a-pna-series-microwave-network-analyzers.pdf
	or
	    https://siglentna.com/wp-content/uploads/dlm_uploads/2017/10/SSA3000X_ProgrammingGuide_PG0703X_E04A.pdf
'''

import array
from heimdallr.base import *
from heimdallr.instrument_control.categories.spectrum_analyzer_ctg import *

class SiglentSSA3000X(SpectrumAnalyzerCtg):
	
	def __init__(self, address:str, log:LogPile):
		super().__init__(address, log, expected_idn="Siglent Technologies,SSA30")
		
		self.trace_lookup = {}
	
	def set_freq_start(self, f_Hz:float):
		self.write(f"SENS:FREQ:STAR {f_Hz} Hz")
	def get_freq_start(self):
		return float(self.query(f"SENS:FREQ:STAR?"))
	
	def set_freq_end(self, f_Hz:float):
		self.write(f"SENS:FREQ:STOP {f_Hz}")
	def get_freq_end(self,):
		return float(self.query(f"SENS:FREQ:STOP?"))
	
	def set_ref_level(self, ref_dBm:float):
		ref_dBm = max(-100, min(ref_dBm, 30))
		if ref_dBm != ref_dBm:
			self.log.error(f"Did not apply command. Instrument limits values from -100 to 30 dBm and this range was violated.")
			return
		self.write(f"DISP:WIND:TRAC:Y:RLEV {ref_dBm} DBM")
	def get_ref_level(self):
		return float(self.query("DISP:WIND:TRAC:Y:RLEV?"))
	
	def set_y_div(self, step_dB:float):
		
		step_dB = max(1, min(step_dB, 20))
		if step_dB != step_dB:
			self.log.error(f"Did not apply command. Instrument limits values from 1 to 20 dB and this range was violated.")
			return
		self.write(f":DISP:WIND:TRAC:Y:SCAL:PDIV {step_dB} DB")
	def get_y_div(self):
		return float(self.query(f":DISP:WIND:TRAC:Y:SCAL:PDIV?"))
	
	def set_res_bandwidth(self, rbw_Hz:float, channel:int=1):
		self.write(f"SENS:BWID:RES {rbw_Hz}")
	def get_res_bandwidth(self, channel:int=1):
		return float(self.query(f"SENS:BWID:RES?"))
	
	def set_continuous_trigger(self, enable:bool):
		self.write(f"INIT:CONT {bool_to_ONFOFF(enable)}")
	def get_continuous_trigger(self):
		return str_to_bool(self.query(f"INIT:CONT?"))
	
	def send_manual_trigger(self):
		self.write(f"INIT:IMM")
	
	def get_trace_data(self, trace:int, use_ascii_transfer:bool=False):
		''' Returns the data of the trace in a standard waveform dict, which
		
		has keys:
			* x: X data list (float)
			* y: Y data list (float)
			* x_units: Units of x-axis
			* y_units: UNits of y-axis
		
		'''
		
		# Make sure trace is in range
		count = int(max(1, min(trace, 3)))
		if count != count:
			self.log.error(f"Did not apply command. Instrument limits values to integers 1-3 and this range was violated.")
			return
		
		# Get Y-unit
		
		
		# Run ASCII transfer if requested
		if use_ascii_transfer:
			
			self.write(f"FORMAT:TRACE:DATA ASCII") # Set format to ASCII
			data_raw = self.query(f"TRACE:DATA? {trace}") # Get raw data
			str_list = data_raw.split(",") # Split at each comma
			del str_list[-1] # Remove last element (newline)
			float_data = [float(x) for x in str_list] # Convert to float
			
		else:
		
			# Set data format - Real 64 binary data - in current Y unit
			self.write(f"FORMAT:TRACE:DATA REAL")
			
			# Read data - ask for data
			self.write(f"TRACE:DATA? {trace}")
			data_raw = []
			while True:
				try:
					byte = self.inst.read_bytes(1)
					data_raw.append(byte)
				except:
					break
			data_raw = data_raw[0]
			
			# Skip first 4 bytes (number of elements) and last byte (newline)
			float_data = list(array.array('f', data_raw[4:-1]))
		
		# Generate time array
		f_list = list(np.linspace(self.get_freq_start(), self.get_freq_end(), len(float_data)))
		
		out_data = {'x':f_list, 'y':float_data, 'x_units':'Hz', 'y_units':'dBm'}
		
		# Convert Y-unit to dBm
		return out_data
		
		# trace_name = self.trace_lookup[trace]
		
		# # Select the specified measurement/trace
		# self.write(f"CALC{channel}:PAR:SEL {trace_name}")
		
		# # Set data format
		# self.write(f"FORM:DATA REAL,64")
		
		# # Query data
		# return self.query(f"CALC{channel}:DATA? SDATA")
		
	
		
	# def set_averaging_enable(self, enable:bool, channel:int=1):
	# 	self.write(f"SENS{channel}:AVER {bool_to_ONFOFF(enable)}")
	# def get_averaging_enable(self, channel:int=1):
	# 	return str_to_bool(self.write(f"SENS{channel}:AVER?"))
	
	# def set_averaging_count(self, count:int, channel:int=1):
	# 	count = int(max(1, min(count, 65536)))
	# 	if count != count:
	# 		self.log.error(f"Did not apply command. Instrument limits values to integers 1-65536 and this range was violated.")
	# 		return
	# 	self.write(f"SENS{channel}:AVER:COUN {count}")
	# def get_averaging_count(self, channel:int=1):
	# 	return int(self.query(f"SENS{channel}:AVER:COUN?"))
	
	# def send_clear_averaging(self, channel:int=1):
	# 	self.write(f"SENS{channel}:AVER:CLE")
	
	# def send_preset(self):
	# 	self.write("SYST:PRES")